#
# <20170115.1846.39>
#
# This code is ment to be attached to a "text" window
# as a menu option (or just plain called after the window is created,)
# It is assumed that that window contains the contents of a file.
# When called, the function "follow" will cause trailing lines 
# of the file to be added to the window as they are written to 
# the file (i.e. this is just like "tail -f")

# This wonder is peformed by calling tail to do the hard work...

# To call
#
#  followFile w file po
#
#     where: w is the "text" window
#            file  is the file displayed in that window
#            endln is the number of lines in the original file
#            po    <optional> is a logical. If true position the window 
#                  such that the last line of text is visable. 
#                  If false, do not.
#
# If "po" is false the window will be positioned only if the cursor is
# at the end of the last line prior to any update that adds new line.  
# In such a case the cursor will remain at the end of the last line.
#
# If the window is empty, the whole file will be written to it as well
# as any new lines as they are written to the file.

# If the "file" is not readable or "tail" is not a program on the system
# the function will return with a string indicating the problem.  Other wise
# an empty string is returned.
proc followFile {w file endln {po 0}} {
  set r [followFile_ $w $file $endln $po]
  if {$r != {}} {
    PopError $r
  }
}

proc followFile_ {w file endln po} {
  global config glob
  # We have several ways to make this work for any number
  # different windows.  What we do is hang everything on the window name.
  # We will release all the globals when the window is closed.
  #
  if {[info exists glob($w,fid)]} {
    return [_ "Follow already active on window %s" $w]
  }
  if {![winfo exists $w]} {
    return [_ "Window %s does not exist." $w]
  }
  if {![file readable $file]} {
    return [_ "File %s is not readable." $file]
  }
  if {[auto_execok [lindex $config(cmd,tail) 0]] == {} } {
    return [_ "Command '$config(cmd,tail)' (config(cmd,tail)) not found"]
  }
  bind $w <Destroy> "+followFileCleanup $w"

  # set ex [format [FixFormatString $config(cmd,tail)] $endln \
  # 	      [FixFileName $file -s 1]] 
  # set cmd [ReSpaceString {} $ex]
  #
  # This is a real hack, but I want to keep the interface.
  # frECF wants to put quotes around all %s but we don't want that here
  # so we format the first %s here...
  #
  set cpart [format "[subst -nob $config(cmd,tail)]" [incr endln] %s]
  frputs "follow:  " cpart file
  set cmd [frECF [list {*}$cpart] [list $file] -f 1]
  set r [catch {open "|[list {*}$cmd 2>@1]" r} fid]
  #set r [catch {open "|$cmd 2>@1" r} fid]
  frputs "opened follow pipe  " cmd 
  if {!$r} {
    fconfigure $fid -buffering none
    fconfigure $fid -blocking 0
    fconfigure $fid -translation auto
    chan event $fid readable "followFileCompletePipe $w $fid"
    set glob($w,fid) $fid
    set glob($w,po) $po
    return {}
  } else {
    return $fid
  }
}

proc followFileCompletePipe { w fid} {
  global glob 
  Log [_ "%s called" followFileCompletePipe]
  set out ""
  set end [$w index "end-1 chars"] 
  set curs [$w index insert]
  set curserAtEnd [expr {$end == $curs}]
  set r [catch {set out [read $fid]} err]
  set eof [eof $fid]
  frputs "followFile  " r out err eof fid end curs
  if {$out != ""} {
    $w insert end "$out"
    if {$curserAtEnd} {
      $w mark set insert end
    }
    if {$curserAtEnd || $glob($w,po)} {
      $w see end
    }
  }
  if {$r || $eof } {
    # do we need this???
    followFileCleanup $w
    if {$eof} {
      Log [_ "Eof found on follow?? Is -f option coded?"]
    } else {
      Log [_ "followFile error %s, aborting follow" $err]
    }
  }
}

proc followFileCleanup {w} {
  global glob
  if {[info exists glob($w,fid)]} {
    catch "close $glob($w,fid)"
    unset glob($w,fid)
    unset glob($w,po)
  }
  return
}
