#
# <20180110.1745.19>
#
# This was shamelessly lifted (-and greately modified) from the
# init.tcl source in tcl.

# But why??
#
# First:   That version returns and saves files with paths of "./"
#           This, of course, fails if the working directory changes.
# Second:  That version returns files that are not executable, if
#          such files exist in a given path dir. Or in PATHEXT
#          I.e. if looking
#          for xyz we expect xyz.exe not xyz even though it may exist.
# Third:   I would contend that files found in "." should not be
#          cached if "." is not in PATH. In fact anything not in Path
#          should not be cached.
# Fourth:  I would also contend that "." should be searched before
#          [info nameofexecutable].
# Fifth:   Failure to find an executable should not be cached. After
#          all the world is still spinning, so it may appear next time
#          if only we look.
# Sixth:   I would contend that system paths not in PATH should be
#          searched only after those in PATH.
# Seventh  Extensions in PATHEXT but not in {.cmd .exe .com .bat} are
#          returned as executable, but are not. These types require
#          an interpretor. We provide an option to supply this.
# Eighth   The usefull shell builtins: assoc, ftype, mkdir, mklink,
#          and move are missing.
#          
# Additional options:
#  -execfor <fun>   Function to call to get executable for a given file ext.
#                   This function is passed the file name and is
#                   expected to return {} or the full path file name
#                   of the executable along with the given file in
#                   a string that can be passed to exec.

# Note that info executable doesn't work under Windows, so we have to
# look for files with .exe, .com, or .bat extensions.  Also, the path
# may be in the Path or PATH environment variables, and path
# components are separated with semicolons, not colons as under Unix.

# None of this is usable unles we are in MS windows...

# We need this outside of the MSW only so that build index will find it.
# We will call this only on windows platforms. It forces replacement
# of the standard auto_execok code...

proc load_now_special_auto_execok {} {}

if {$tcl_platform(platform) eq "windows"} {


  # we put this here to share it...
  namespace eval auto_execNS {
    variable path {}
  }
  #
  proc auto_execok {name args} {
    global auto_execs env tcl_platform
    upvar ::auto_execNS::path path

    # Just in case we have a quote protected name...
    set name [regsub {\"([^\"]*)\"} $name {\1}]
    if {[info exists auto_execs($name)]} {
      return $auto_execs($name)
    }

    set shellBuiltins [list assoc cls copy date del erase dir echo\
			   ftype mkdir md mklink move rename ren rmdir \
			   rd time type ver vol]
    if {$tcl_platform(os) eq "Windows NT"} {
      # NT includes the 'start' built-in
      lappend shellBuiltins "start"
    }
    if {[info exists env(PATHEXT)]} {
      # Add an initial ; to have the {} extension check first.
      set execExtensions [split [string toupper ";$env(PATHEXT)"] ";"]
    } else {
      set execExtensions [list {} .COM .EXE .BAT .CMD]
    }
    if {[lsearch -exact -nocase $shellBuiltins $name] != -1} {
      # When this is command.com for some reason on Win2K, Tcl won't
      # exec it unless the case is right, which this corrects.  COMSPEC
      # may not point to a real file, so do the check.

      # We do recursion for this...
      
      if {[set cmd [auto_execok $env(COMSPEC)]] != {}} {
	return [set auto_execs($name) [list $cmd /c $name]]
      }
    }
    # build the full path early so we can easily check if the
    # answer is from a dir on the path.

    # order the path so that we check the right stuff first
    # also split now so we can use lsearch on it.
    
    # As an aside we note that windows works with two paths, one
    # for the user and one for the system. The problem, IMHO, is that
    # when it builds the PATH env it puts the systems path first.
    # While we contend this is wrong, it is not really possible to
    # fix it here.
    
    foreach var {PATH Path path} {
      if {[info exists env($var)]} {
	lappend pathT {*}[split $env($var) ";"]
      }
    }

    # now add the tcl stuff
    
    lappend pathT [file dirname [info nameof]]
    # Now the windows stuff

    if {[info exists env(WINDIR)]} {
      set windir $env(WINDIR)
      if {$tcl_platform(os) eq "Windows NT"} {
	lappend pathT $windir/system32
      }
      lappend pathT $windir/system $windir
    }
    
    foreach ph $pathT {
      if {$ph ni $path && $ph != {}} {
	lappend path $ph
      }
    }
    
    # Now if a path was provided...
    set fileNorm [file normalize $name]
    if {[llength [file split $name]] > 1} {
      return [auto_checkIfEx $fileNorm]
    }

    # puts "now the path $path"
    # A simple name with no path...
    # lh $path
    foreach dir [concat {.} $path] {
      # Skip already checked directories
      if {[info exists checked($dir)] || ($dir eq "")} {
	continue
      }
      set checked($dir) {}
      if {[set rtn [auto_checkIfEx [file normalize $dir]/$name]] != {}} {
	return $rtn
      }
    }
    # puts "[array get checked]"
    return {}
  }

  # This is a local function to consolidate common code.
  # Do not attmept to call this else where..

  proc auto_checkIfEx {fileNorm} {
    global env auto_execs
    upvar execExtensions execExtensions
    upvar path path
    upvar name name
    upvar args args
    frputs fileNorm
    foreach ext $execExtensions {
      set file ${fileNorm}${ext}
      if {[file exists $file] && ![file isdirectory $file]} {
	# make sure this is executable
	if {[set thisext [string toupper [file extension $file]]] == {} ||\
	      $thisext ni $execExtensions} {
	  continue
	}
	if {$thisext in {.COM .EXE .BAT .CMD}} {
	  set rtn [list $file]
	} else {
	  # extension exists and is not in executable class
	  # so it must be a recognizable extended "ext"
	  lassign [concat {*}$args] op val
	  set rtn [list [auto_execok $env(COMSPEC)] /c $file]
	  if {[string range $op 0 1] == "-e"} {
	    if {[set help [eval [list $val $file]]] != {}} {
	      set rtn $help
	    }
	  }
	}
	# path names are native
	if {[lsearch -nocase -exact $path \
		 [file native [file dirname $file]]] != -1} {
	  set auto_execs($name) $rtn
	}
	return $rtn
      }
    }
    return {}
  }
}

# For what its worth, I am not too keen on the *nix version of auto_execok
# as it has some of the same flaws.  However, at this time I just don't
# use it.  
