#
#<20190707.1737.32>
#
# This module contains bits of filerunner that are pure windows
# Putting this stuff here lightens the load on other platforms
#
# This bit will be sourced in global space

# This function (by name) triggers sourceing by the pkgIndex for
# the platform
lappend ::startTimes [list [clock milliseconds] "First light in Win Platform"]

if {[set it [info script]] ni $::auto_path} {
  lappend ::auto_path $it
}
unset -nocomplain it

proc load_now_fr_bits {} {}

# This is the turn on setup unique to this platform

set glob(os) WIN32
set ::MSW 1
set ::CASEops "-nocase"
set ::TMP $::env(TMP)

proc buildTbarIcon {} {
  global glob
  wm iconbitmap . -default $glob(lib_fr)/icons/filerunner-icons.ico
  # This is called after init config so we have an opportunity do
  # platform specific config initialization.
  # We may want to rethink this later....
  # For now we will put these at the head of the button list
  
  # Here we add platform dependent buttons for the middle list.
  # format is the same as in fr.
  set properties {
    {{Properties} CmdProperties p 0 \
	 {[_b "Change the mode flags for selected file(s)." ]}}
    {{Short Cut} CmdCreateShortCut p 0\
	 {[_b "Create a short cut for selected file(s)."]}}
    {{Short CutAS} CmdCreateShortCutAs p 0\
	 {[_b "Create a short cut for selected file(s)."]}}
  }
  foreach platformButton $properties {
    lappend add [list [lindex $platformButton 0]]
    if {[lsearch -exact -index 0 $glob(cmds,list) [lindex $platformButton 0]] == -1} {
      lappend glob(cmds,list) $platformButton
    }
  }
  set ::config(middle_button_list) [linsert $::config(middle_button_list) 0 {*}$add]
}


lappend ::startTimes [list [clock milliseconds] "After icon set up"]

# if this is a wrapped process we move the twapi package to our
# private location to, well basically uncompress it.
if {[namespace exists freewrap]} {
  # If it exists and is the right version, leave well enough alone
  # Note that version is in the pkgIndex so we just compare the whole thing
  set r [catch {read [set fid [open /f/filerunner/packages/twapi/pkgIndex.tcl r]]} dat]
  set r2 [catch {read [set fid2 [open $glob(conf_dir)/twapi/pkgIndex.tcl r]]} dat2]
  catch {close $fid}
  catch {close $fid2}
  if {$r + $r2 != 0 || $dat != $dat2} {
    file copy -force [zvfs::list */twapi] $glob(conf_dir)/twapi
  } else {
    # puts "twapi ok"
  }
  lappend ::auto_path $glob(conf_dir)/twapi
}

# puts "$::auto_path"
package require twapi
set ::Elevated [expr {[twapi::get_process_elevation] == "full"}]

lappend ::startTimes [list [clock milliseconds] "After Twapi set up"]

# We are agressive about cleaning up our tmp dir
proc cleanTmpFiles {} {
  global glob
  package require dde
  dde servername [pid]
  set servers [dde services TCLEval {}]
  foreach pid $servers {
    lappend pids [lindex $pid 1]
  }
  set dmess {}
  set glob(tmpdir) $glob(conf_dir)/tmp[pid]
  if {![file isdirectory $glob(tmpdir)]} {
    if {![file exists $glob(tmpdir)]} {
      set r [Try { file mkdir $glob(tmpdir) } "" 1]
      if {$r != 0} {
	lappend dmess [_ "Can't create %s. May be problems later" $glob(tmpdir)]
      }
    }
  }
  foreach dir [glob -tails -directory $glob(conf_dir) tmp*] {
    set who [regsub {tmp} $dir {}]
    if {$who == {} && [glob -nocomplain $glob(conf_dir)/tmp/*] != {}} {continue}
    if {$who ni $pids} {
      if {[catch {file delete -force $glob(conf_dir)/tmp$who} ms]} {
	lappend dmess $ms
      }
    }
  }
  if {$dmess != ""} {
    lappend ::mess "Error managing tmp files:"
    lappend ::mess {*}$dmess
  }
}
########################## end of general platform set up ##################
proc inotifySetUp {} {

}
# stubs for windows, used only in unix

proc GetUidGidString { file uid gid } {
    return ""
}

proc GetStringFromMode { mode } {
  global glob
  if { $::tcl_platform(platform) == "windows"} {
    return [format "0x%x" $mode]
  }
}
# This routine is to get the full disc size. On
# *nix system it also returns the mount point of the
# containing file system


# Ok, take 2. Lets use the twapi com code to do this.
# our private global
namespace eval ::oo {
  variable fso  {}
  variable drives {}
}

proc GetDF { dir } {
  # we need the letter and the : only...
  set volId [string range $dir 0 [string first ":" $dir]]
  if {$volId == {}} {return "??"}
  if {[catch {[$::oo::drives Item $volId] FreeSpace} ans] != 0} {
    catch {$::oo::fso    -destroy}
    catch {$::oo::drives -destroy}
    set ::oo::fso [::twapi::comobj Scripting.FileSystemObject]
    set ::oo::drives [$::oo::fso Drives]
    if {[catch {[$::oo::drives Item $volId] FreeSpace} ans] !=0} {return "??"}
  }
  #frputs ans
  if {[string index $ans end-1] == "."} {
    set ans [string range $ans 0 end-2]
  }
  #frputs ans
  return [list [formatDF $ans] {}]  
}


# This code in some senses emulates the linux version of inotify to
# handle the same functionallity on MSW.

# The watch takes:
# what   add or remove to set up or remove a watch
# which  the directory to monitor
# how    the watch flags

proc notifywatch {what which {how {}}} {
  global ::inotify::dirs
  # We only allow two monitors at a time, attempting more is
  # An error..
  if {$what == "add"} {
    if {[llength $dirs] >=2} {
      frputs dirs which how
      return -code error "Attempting to monitor more than 2 dirs"
    }
    frputs which how
    if {[catch {::twapi::begin_filesystem_monitor \
		    $which [list notecallback $which] {*}$how} out] != 0} {
      frputs out which how
      return -code error $out
    }
    # sweet success...
    lappend dirs [list $which $out]
    return $which
  }
  # must be a remove
  set ndirs {}
  foreach dir $dirs {
    lassign $dir di han
    frputs dir di han
    if {$di == $which} {
      ::twapi::cancel_filesystem_monitor $han
      incr found
      continue
    }
    lappend ndirs $dir
  }
  if {![info exists found]} {
    frputs  which dirs ndirs
    return -code error "$which not found in monitored directories"
  }
  set dirs $ndirs
}

proc notecallback {args} {
  global glob config
  set did 0
  set found {}
  frputs args
  # For windows notecallback gets:
  # arg0   the directory we are monitoring
  # arg1   the monitor handle (not of any use here)
  # arg3   a list of dublets (where fd is the file name):
  # 'added' fd, 'removed' fd, 'modified' fd, 'renameold' fd, 'renamenew' fd
  # or 'error' error-code.
  # Of these we only seperate modified and error...
  lassign $args dir han dublets
  foreach lr {left right} {
    if {$glob(notify_id,$lr) == $dir} {
      foreach {type file} $dublets {
	set key [string index $type 0]
	if {$key == "m"} {
	  lappend glob(n_files,$lr) $file
	} elseif {$key != "e"} {
	  lappend found $lr
	  if {$key == "a" && [capture $ou] == 1 } {
	    proc capture { note } { }
	  }
	} else {
	  # We have an error, shut down the whole thing
	  notifywatch remove $dir
	  return -code error "Directory modify monitor error $file"
	}
      }
    }
  }
  doInotifyUpdate $found
}
############################### End of notify code ################
############################### build Tree stub    ################
proc buildTree {args} {}
############################### End of Tree stub   ################

############################### displays setup     ################
# Here we attempt to gather info on attached displays
# This code depends on xrandr being available to query the display system
#
namespace eval ::displays {
  # monitors is set at config time... don't change it
  variable monitors
}
proc ::displays::init {} {
  # First get the borders
  lassign [twapi::get_window_coordinates [list [wm frame .] HWND]] xmin ymin xmax ymax
  lassign [split [wm geo .] x+] width height
  set ::displays::monitors(width)  [expr {$xmax - $xmin - $width }]
  set ::displays::monitors(height) [expr {$ymax - $ymin - $height}]
  set ::displays::monitors(monitors) {}
  foreach display [twapi::get_multiple_display_monitor_info] {
    array set tmp $display
    lassign $tmp(-workarea) xmin ymin xmax ymax
    lappend ::displays::monitors(monitors) [list $xmin $xmax $ymin $ymax]
    array unset tmp
  }
  Log "Monitors found: [array get ::displays::monitors]"
  return 
}
# For MSW we do this now. It is fast and this means we don't have to save it
# in the config file
::displays::init
############################### end of displays setup ################


# fileNameTranslate returns a list of doublets in a format suitable for 
# 'array set'..  The following will only be present if needed:
#
# scheme {file nfs cifs}
# path   for file -$sfile, for nfs target file, for cifs part of $sfile to add
#                                                        to share location
# host   for file not needed, for nfs|cifs host ip 
# user   for file not needed, for nfs|cifs user used in fstab
#
# share  only for nfs & cifs the share name (host:/??) or (//host/???)
# If a vaue is unknown it will be {}.
proc fileNameTranslate {sfile} {
  set localLetter [string range $sfile 0 1]
  # This will throw an error for a local drive
  if {[catch {twapi::get_client_share_info\
		  $localLetter\
		  -remoteshare\
		  -user} shareData] == 0} {
    set scheme "cifs"
    array set tmp $shareData
    set share [regsub -all {\\} $tmp(-remoteshare) {/}]
    return [list path [string range $sfile 1 end]\
		share  $share\
		host $hostname \
		user tmp(-user)\
		scheme cifs\
		pwd cifs://$tmp(-user)@[string range $share 2 end]]
  } else {
    return [list path $sfile scheme file]
  }
}

# Here we have a couple of entries for the Utilities menu
#
namespace eval ::UtilsMenu {
  # These entries are added to the utilities menu for Windows
  variable ents {
    {-label {Find remote shares} -command {getRemoteShares}}
  }
}

#     {-label {Assign drive letter} -command {setWinVolLetter}}

proc setWinVolLetter {} {
}

proc getRemoteShares {} {
  Try {PopWarn \
           "This may take a while. You may continue doing other things\
           \n while we work. If new shares are found you will be presented\
           \n with options on how to handle them and, if required user/password\
           \n information.\
           \n\nWhen done the \"async\" count will go to zero\
           \n and found shares will be in the <volume> list."} -a
  realWaitForIdle
  Try {::UtilsMenu::findShares} -a
}

proc ::UtilsMenu::findShares {} {
  set save 0
  set foundShares [::findUNC::getShares]
  frputs foundShares
  puts "$foundShares"
  set connectedShares [regsub -all {\\} [twapi::wnet_connected_resources] {/}]
  frputs foundShares connectedShares
  foreach {share drive} $foundShares {
    # see if the share is connected
    frputs share drive
    set pSlash [string range $share 2 end]
    #  set pwe [lsearch -glob  -index 0 $::pwLocker::pwAddress "cifs://*@$pSlash"]
    set r [catch {pwLocker::getPw cifs://*@$pSlash noprompt 1} rs]
    set loc [lsearch -exact -nocase -index 1 $connectedShares $share]
    if {[llength $loc] == 0 || $r != 0 || $rs == -1} {
      # a new share, lets set it up....
      # Test if we need a name/password
      set noPw "a"
      if {$loc != {}} {
	# not connected, try now
	set result [catch {twapi::connect_share [regsub -all {/} $share {\\}]} ops]
	if {[regsub -all {\\} $ops {/}] == $share} {
	  # all is good, disconnect and ask
	  set noPw "no"
	  catch {twapi::disconnect_share [regsub -all {/} $share {\\}]}
	}
      }
      # if share is already in the locker, we leave it alone.
      # still we are aware that it may be there with bad credentials.
      
      # see if already in the pwLocker
      if {$rs == -1} {
	foreach vol [file volume] {
	  lappend volInUse [string range $vol 0 1]
	}
	set disList {}
	if {$drive == {}} {
	  set dMess "No drive letter has been assigned. Fill in \"drive\" below\
                    \nwith any but the used drive letters ("
	  append dMess $volInUse ")\n or \"*\" to let the system assign a drive letter."
	  lappend disList "driveletter" {}
	} else {
	  set dMess "A drive letter has been assigned ("
	  append dMess $drive "). This can not be changed."
	}
	set pstring "\nThis share is not in the password locker. \n"
	append pstring "It requires " $noPw " user/password to be access.\n"\
	    "To dismiss and not set this share up click on \"Cancel\" or the top right X\n"\
	    $dMess "\n"\
	    "The flag field indicates if filerunner should not generate a Display\n"\
	    "name for this share (the default is to generate a Display Name). Meanings:\n"\
	    "No    do not generate a Display Name.\n"\
	    "Hot   put the display name in the HotList only.\n"\
	    "Both  put the Display Name in both the \"volume\" and HotList.\n"
	lappend disList "flag" "Both"
	set error {}
	set pLock(password) {}
	while {1} {
	  set r [catch {pwLocker::getPw cifs://$pSlash\
			    prompt [list $pstring $error]\
			    display $disList\
			    startwith $pLock(password)\
			    returnall 1} an]
	  frputs an
	  if {$r != 0 || $an == -1} {
	    # We are not to enter this share.
	    break
	  }
	  array set pLock $an
	  # normalize a few things
	  set rqDrive {}
	  if {[info exist pLock(driveletter)]} {
	    set pdrive [string toupper [string index $pLock(driveletter) 0]]:
	    if {$pdrive == ":"} {
	      # No drive letter...
	      set pdrive {}
	    }
	    frputs pdrive volInUse
	    if {$pdrive == "*:"} {
	      set pdrive "*"
	    } elseif {$pdrive in $volInUse} {
	      set error "\nDrive letter ("
	      append error $pdrive ") already in use."
	      continue
	    }
	  } else {
	    set pdrive $drive
	  }
	  if {$pdrive != {}} {
	    set rqDrive [list -localdevice $pdrive]
	  }
	  if {$drive != {}} {
	    # restating the drive letter here is not allowed?
	    set rqDrive {}
	  }
	  # Verify we can connect with this name/password
	  set r [catch {twapi::connect_share [regsub -all {/} $share {\\}]\
			    -user  $pLock(user)\
			    -password [twapi::conceal $pLock(password)] \
			    {*}$rqDrive\
			  } error]
	  frputs r error
	  if {$r != 0} {continue}
	  if {$pdrive == "*"} {
	    set pdrive $error
	  }
	  # Things look good.
	  # Note we always save, but if the don't save flag is set
	  # we null the password first.
	  if {!$pLock(keep)} {
	    set pLock(password) {}
	  }
	  pwLocker::putPw cifs://$pLock(user)@$pSlash $pLock(password) \
	      $pdrive $pLock(flag)
	  incr save
	  break
	}
      }
    }
  }
  if {$save != 0} {
    passWordToDN
    SaveConfig
  }
}
