#
# <20190411.1757.32>
#
# Here is a routine to eliminate symbolic links in a path
# In the *nix world pwd is the way. In MSW it is not so simple
#
# proc getPath {dir} {
#   set parts [file split [file norm $dir]]
#   set name {}
#   foreach part $parts {
#     set name [file join $name $part]
#     if {[catch {file readlink $name} an] != 0} {continue}
#     set name [getPath $an]
#   }
#   return $name
# }
# Windows get dir info using windows DIR command
# integer min and max for a list...
namespace eval getWinDir {
  variable timeSize
  variable ownerSize
  variable sizeTypeSize
  variable fileStart
}

# The init code sets up the char positions in the 'dir' strings
# The issue with the date is that it depends on what how the  user has set the
# time and date format. Here we check and set up the conversion format
# and the width to expect in the dir listing.

proc ::getWinDir::dateWidth {} {
  array set sysDateFormat [::twapi::get_locale_info \
			       [twapi::get_system_default_lcid] \
			       -stimeformat \
			       -sshortdate]
  if {[string index $sysDateFormat(-stimeformat) end] == "t"} {
    set todWidth 8
  } else {
    set todWidth 5
  }
  switch -exact $sysDateFormat(-sshortdate) {
    M/d/yy     -
    MM/dd/yy   -
    yy/mm/dd   {set dateWidth 8}
    M/d/yyyy   -
    MM/dd/yyyy -
    yyyy-mm-dd {set dateWidth 10}
    dd-MMM-yy  {set dateWidth 9}
    default {}
  }
  return  [expr {$dateWidth + $todWidth + 2}]
}

proc formatDF {free} {
  regsub -all {[^0-9]} $free "" free
  # should now be a pure number.  If nil, use zero...
  # better than failing in the following code...
  if { $free == "" } {
    set free 0
  }
  set indx 0
  while { $free > 1024. } {
    incr indx
    set free [expr {$free / 1024.}]
  }
  #  puts "$indx $free"
  set suf { " " K M G T}
  return [format {%.3f%s} $free [lindex $suf $indx]]
}


proc ::getWinDir::init {} {
  foreach var {nameOffset ownerOffset} {
    variable $var
  }
  # we create a temp file in the home directory.  We know its name
  # and size so we should be able to find the rest. The over all
  # format is date size or <flag> owner, name & possible link
  # date is "mm/dd/yyyy hh:mm am/pm"
  #
  set pwd [pwd]
  cd [file dir $::env(APPDATA)]
  set target [file tail $::env(APPDATA)]
  set dirCmd [list cmd /u /c dir /a /q /-c /a $target*]
  set r [catch "exec {*}$dirCmd" dirList]
  cd $pwd
  set dirList [encoding convertfrom unicode $dirList]
  # Here is what we want:
  # date-time|     <??> or size      |owner        |file name & link
  # dateLn   ^ enddir ^  ownerOffset ^  nameOffset ^
  # Where the actual location of the xxOffset is offset from the dateLn
  # thus 'dateLn' + ownerOffset gives the owner location etc.
  # now find AppData which we know is a directory name"
  set nameAbs [string first $target $dirList]
  set start [string last \n  $dirList  $nameAbs]
  incr start
  set dateLn [dateWidth]
  set nameOffset [expr {$nameAbs - $start - $dateLn}]
  set tmpString [string range $dirList [string first > $dirList $start]+1 $nameAbs-1]
  # tmpString is from '>' to name. In there is the owner
  # we strip leading blanks leaving the owner field whos length
  # gives us the ownerOffset 
  set ownerOffset [expr {$nameOffset -\
			     [string length [string trimleft $tmpString]]}]      
  # frputs nameAbs start dateLn nameOffset ownerOffset  tmpString 
  
  
}

::getWinDir::init

proc getWindowsDIRfilelist {lastmtime dir showall showDirs showSort} {
  global glob config  
  global ::getWinDir::ownerOffset
  global ::getWinDir::nameOffset

  upvar 1 $lastmtime mylastmtime
  upvar dl dl sortlist sortlist
  upvar 1 inst inst
  set dateLn [::getWinDir::dateWidth]
  cdAndTest $dir/
  set mylastmtime [file mtime $dir]
  # by using the "/u" we get a unicode version...
  set dirCmd "cmd /u /c dir /a /q /-c /a[expr {$showall ? {} : "-h"}]"
  set dirList {}
  set r [catch "exec {*}$dirCmd" dirList]
  # convert the unicode to utf-8
  set dirList [encoding convertfrom unicode $dirList]
  set nl [split $dirList \n]
  # Ok, now we go through the list and pluck out what is needed.
  # do the invariant stuff
  set sortval_d [switch -exact $showDirs {
    dirsfirst {expr {0}}
    mixed     {expr {1}}
    dirslast  {expr {2}}
  }]
  set needAtime [expr {$showSort == "atime" || \
			   $showSort == "ratime" || \
			   [lsearch -index 0 -exact \
				$config(ListBoxColumns,$inst) "atime"] != -1}]
  set atime {}
  set ctime {}
  set shortName {}
  # This reg expression is used to seperate the links from file names.
  # It also removes \??\ flags from the link paths
  # set regPat  {([^\[]*) [\[]([^\]]*)} ; # This line is ok (but does it do the job)
  set regPat  {([^[]*) [[](\\\?\?\\)?([^]]*)} ;# This line throws emacs indention code, sigh!
    
    foreach ent $nl {
      if {$ent == {}} {continue}
      if {[string index $ent 0] == " " && \
	      [string length $ent] < ($dateLn + $nameOffset) &&\
	      [string match -nocase "* bytes free" $ent]} {
	# This is the free bytes line
	# isolate the number
	set num [string range $ent 0 [string first "bytes" [string tolower $ent]]-1]
	set glob($inst,df) [formatDF [lindex $num end]]
	frputs num glob($inst,df)
	continue
      }
      if {[string index $ent 0] == " "} {continue}
      # add the offsets date length to each offset.
      # better yet, get rid of the date, we don't use it anyway
      set ent      [string range $ent $dateLn end]
      set file     [string trim [string range $ent $nameOffset end]]
      set owner    [string trim [string range $ent $ownerOffset $nameOffset-1]]
      set sizeType [string trim [string range $ent 0 $ownerOffset-1]]
      # frputs ent file owner sizeType
      
      # that is all that we have, derive the rest
      # set mtime [clock scan $mtime]
      set size 0
      set filetype n
      set linkname {}
      # frputs file  sizeType
      switch -exact -nocase $sizeType {
	"<dir>"      {set filetype d}
	"<symlinkd>"  -
	"<junction>" {
	  set filetype ld
	  regexp $regPat $file n file fo linkname
	  # frputs file linkname fo
	}
	"<symlink>" {
	  set filetype l
	  # frputs file linkname fo
	  regexp $regPat $file n file fo linkname
	}
	default    {
	  set size $sizeType
	  if {[file extension $file] == ".lnk"} {
	    if {[catch {LnkFile ./$file to dir} out] == 0 && $out} {
	      set filetype [expr {$dir ? {wld} : {wl}}]
	      set linkname $to
	    }	  
	  }

	}
      }
      if {$needAtime} {
	set atime [file atime $file]
      }
      set mtime [file mtime $file]	    
      # no ctime (hard access on windows)
      # if {$needCtime} {
      #   set ctime [file ctime $file]
      # }
      set tmp [switch -exact $showSort {
	atime     {format "%011d" $atime }
	ctime     {format "%011d" $ctime }
	mtime     {format "%011d" $mtime }
	ratime    {format "%011d" [expr 2147483647 - $atime]}
	rctime    {format "%011d" [expr 2147483647 - $ctime]}
	rmtime    {format "%011d" [expr 2147483647 - $mtime]}
	size      {format "%011d" $size }
	extension {file extension $file}
	default   {expr { " "}}
      }]
      switch -exact $filetype {
	wl -
	n  -
	l  {set sortval 1$tmp}
	d   -
	wld -
	ld  {set sortval $sortval_d$tmp}
      }
      lappend dl [list $sortval $file $filetype $size \
		      $mtime "" \
		      $owner \
		      $linkname \
		      $shortName\
		      $atime $ctime]
      # It would appear that sort has a problem with messy file names..
      lappend sortlist "$sortval [list $file]"
    } ; # foreach
    
  }
