#
# <20190406.2019.43>
#
# *.lnk resolution for windows
#
# For windows we will use twapi to decode the *.lnk
#
# This code decodes the windows .lnk file.  It provides several bits of
# information about the link as defined below.
#
# Our interface:
# Return value is 1 if "filename" is a lnk file. 
# Otherwise 0 (includes unreadable not a filename, etc.)
# If 0 is returned and "to" is provided, "to" will be set to a reason.
# "to" if provided is a var that will be set to the file the link points to.
# "dir" if provided will be set to 1 if the link points at a directory, else 0.
# "try" if 1, will attempt to decode even if the "required fields" are wrong.
#       The required fields are "header length == 72" and "clid is correct"
# "diag" Causes debug info to be spilled out to stdout

proc LnkFile {filename {to {}} {dir {}} {try 0} {diag 0}} {
  set r [catch {twapi::read_shortcut $filename \
		    {*}[expr {$try ? "-noui" : "-noui -nosearch"}]} ans]
  set r [expr {!$r}]
  if {$r} {
    array set info $ans
    set ans [regsub -all {\\} $info(-path) {/}]
    if {$dir != {}} {
      uplevel 1 "set $dir [file isdir $ans]"
    }
  }
  if {$to != {}} {
    uplevel 1 "set $to \"$ans\""
  }
  return $r
}
#
# Here we try to verify a lnk pointer for windows
#
# we assume that the C:/ or other such ?:/ refer to the 
# mount point on the system.
# First we try the current volume. If sucessful it is entered in 
# a table for the next time.
# If that fails, we try a little table we keep to translate these.
# If that finds a file, we enter the new translation in our
# table.  If not, we fail with a null name.
#
# The other entry we decode is <??Folder> where ?? appears to be
# the current folder name.  We translate this to "."
#
# To do our work we require the mountpoint of the place the link
# was found and the found link.  We assume the link will be used
# within the current "pwd" (to make the "." work).

#
proc TranslateLnk {link curMountPt} {
  global glob
  set filePrefix {}
  switch -glob $link {
    {?:/*} {
      set first [string toupper [string range $link 0 0]]
      set glob(volume,$first) [file volumes]
      set file $link
    }
    {//*} {
      # Lets assume that it is mounted with the same name
      set file [string range $link 1 end]
    }
    {*.url}   {return $link}
    {./*}     {set file $link }
    default   {set file "./$link" }
  }
  # Check if we found the file...
  if {![file exists $file]} {
    # if this is a ?:/ file and we have a previous working volume, try that
    if {$filePrefix != {} && [info exists glob(volume,$first)]} {
      foreach vol $glob(volume,$first) {     
	if {[file exists [set file $vol/[string range $link 3 end]]]} {
	  return $file
	}
      }
    }
    return {}
  } elseif {$filePrefix != {}} {
    # Allow several prefix to volume translations but only
    # one copy of a given transation.
    if {![info exists glob(volume,$first)] || \
	    $filePrefix ni $glob(volume,$first)} {
      lappend glob(volume,$first) $filePrefix
    }
  }
  return $file
}
