#
# <20190519.1657.44>
#
# This is an expansion on auto_execok for windows.  It takes advantage of
# the possibility that there are some *nix utilities that might be
# the right thing to do.
#
# As in auto_execok we are passed a name and are to look for an appropiate
# way to execute it. In addition, we accept a list of extensions and how
# to run them. Each entry in this list should look like:
#
# <command> {exts}    where exts are a set of glob patterns that a file name
#                     must match to be passed to <command>
# For generality we will recognize a "%s" in <command> as where the file name
# should be.
# We search in this list first...
#
# There are three possible returns here:
# 1) nil, could not find an executable
# 2) Full path to executable
# 3) Full path to an executable that needs the original file inserted 
#    this version returns a value containing "%s" and is a list of two
#    values, the execute part and the modified file that we started from.
#    These will need to be split by the caller for passage to frECF
#  
# In general, we assume that the path to executables may be *nix style
# and others need to be native, we do not attempt the insertion here
# but leave it to the caller.

# All of this is windows only.  If by some odd mistake it get loaded
# else where well, just don't let it do anything.

if {$tcl_platform(platform) != "windows"} {
  # This should terminate the "source" command and thus not load the following
  return
}


proc windowsAutoExecOk {name {extensions {}}} {
  global auto_execs
  upvar ::auto_execNS::path path
  #
  #
  if {[info exists auto_execs($name)]} {
    return $auto_execs($name)
  }
  set ans {}
  foreach ent $extensions {
    lassign $ent cmd pats
    foreach pat $pats {
      frputs "matching " pats pat name extensions
      if {[string match -nocase $pat $name]} {
	set ans $cmd
	break
      }
    }
    if {$ans != {}} {
      break
    }
  }
  if {$ans != {}} {
    # If we have an answer, make sure its pretty...
    # The first entry and each entry after a pipe notation
    # should be a findable executable. Just for grins we
    # will go recursive here to allow any thing we find to
    # be any one of the constructs we look for.
    set start 0
    set out {}
    frputs "try for list.. " ans
    while {1} {
      if {![string is wordchar -strict -failindex indx \
		[string range $ans $start end]] &&\
	      $indx != 0} {
	set indx [expr {$start + $indx}]
	set expand [auto_execok [string range $ans $start $indx-1]]
	if {$expand == {}} {
	  
	  return {}
	}
	# not sure about expanding this...
	append out $expand
	set nx [string first "|" $ans $indx]
	if {$nx == -1} {
	  # we are done...
	  append out [string range $ans $indx end]
	  break
	}
	# a pipe, another name to expand...
	append out [string range $ans $indx $nx]
	while {[string index $ans [incr nx]] == " "} {
	  append out " "
	}
	
	set start $nx
	continue
      }
      # looked like there should have been a program name here, but???
      append out [string range $ans $start end]
      break
    }
    if {[string first "%s" $out] == -1} {
      append out " %s"
    }
    # It does not seem fruitful to cache these
    return [list {*}$out]
  }
  
  if {[set ans [auto_execok $name -e {get_MSW_openCommand 1}]] != {}} {
    frputs "auto_execok returned " ans
    return $ans
  }

  # So we know it is not an executable. Lets see if we have a registry
  # entry for this type... but first it must be a real file..

  
  # Ok, we are desperate now..
  # none of the extension possibilities worked.  We could use *nix file here
  # but it seems faster to just peak at the first line...
  # The assumption at this point is that we have a script that we need to
  # connect with the right program, which should be in the
  # first line of the file (*nix style)

  # We also assume that the path name in such a line was set for *nix and
  # may not be valid here.
  
  # First we need to find the actual file...
  #
  # frputs  name
  if {![file exists [set tfile [file normalize $name]]] ||\
	  [file isdirectory $tfile] ||\
	  ![file readable $tfile]} {
    # We are using the path set up by my_auto_execok replacement
    # It is a list with a lot of redundant entries and it is in
    # native name mode which we change with "file normalize"
    foreach dir $path {
      if {[info exists checked($dir)] || ($dir eq "")} {
	continue
      }
      # frputs "checkdir  " dir
      set checked($dir) {}
      if {[file exists [set tfile [file normalize $dir]/$name]] && \
	      ![file isdirectory $tfile] &&\
	      [file readable $tfile]} {
	frputs "Found  " tfile
	set file $tfile
	break
      }
    }
  } else {
    set file $tfile 
  }
  frputs  file tfile
  if {[info exists file]} {
    set fid [open $file r]
    frputs fid file
    set str [read $fid 200]
    close $fid
    set str [lindex [split $str \n] 0]
    frputs str
    if {[string range $str 0 1] == "#!"} {
      # so far so good, see what the file name looks like
      # We get radical here and look at the tail of the line first
      # The notion is that we may have something like:
      #!/bin/sh -c foo -op
      # where what we want is 'foo -op' as long as 'foo' is an exectuable
      # The prime motivation here is to prevent multiple interpitations
      # of a possibly quoted file name
      set str " [string trim [string range $str 2 end]]"
      set es [string length $str]
      incr es -1
      set tail {}
      set out {}
      while {[set star [string last " " $str $es]] != -1} {
	set fname [string range $str $star+1 $es]
	# Is this a recognizable executable?
	if {$fname != { } && \
		[set out [auto_execok $fname]] != {} || \
		([set tfname [file tail $fname]] != $fname && \
		     [set out [auto_execok $tfname]] != {}) } {
	  break
	}
	# not this, try backing up
	set tail "$fname $tail"
	set es [incr es -1]
      }
      if {$out != {}} {
	# frputs "We have something  " out
	# Still here.. "out" is the result, need to include the rest of the line
	set out [list $out {*}[string range $str $es+1 end] {%s}]
	if {[lsearch -nocase -exact $path \
		 [file native [file dirname $file]]] != -1} {
	  # frputs out file
	  set auto_execs($name) [list $out $file]
	}
	# We need to include the file that got us here as a unit
	# frputs out file
	return [list $out $file]
      } else {
	# ok, not a *nix sort of scritp
	# last ditch, does windows have a registered open for it
      }
    }
    return [list [get_MSW_openCommand 0 $file] $file]
  }
}

# 
# This function uses the file ext to look in the MS windows registry and
# find the command to open the file.  It then builds that command and
# returns the result. The program to run will be in forward / mode
# while the file argument will be in 'native' mode. Null return
# indicates that there is no such entry in the registry.
# If "yes" is false, "%s" will be inserted instead of the file name.
# We will ALWAYS insert one or the other...

proc get_MSW_openCommand {yes file} {
  global env
  upvar ::auto_exts auto_exts

  set ext [file extension $file]
  if {$yes} {
    set file [file native $file]
  } else {
    set file "%s"
  }
  if {![info exists auto_exts($ext)]} {
    # look in rgeistry for the answer..
    set root HKEY_CLASSES_root
    set r [catch [list registry get HKEY_CLASSES_ROOT\\$ext {}] type]
    if {$r != 0} {return {}}
    set r [catch [list registry get \
		      HKEY_CLASSES_ROOT\\$type\\Shell\\open\\command {}] f]
    if {$r != 0} {return {}}
    # To avoid backslash hell, we use forward here...
    set f [regsub -all "\\\\" $f {/}]
    set auto_exts($ext) $f
    # frputs "from reg after bs  " f type
  } else {
    set f $auto_exts($ext)
    # frputs "cache!!  " f
  }
  # now we need to remove any env vars.  They look like %stuff%
  # There are also parm substitution that look like:
  # since we are doing this for "frECF" we will do that convention
  # %0 or %1   takes the target file --> fixed here
  # %2 - %9    gets parms 1 through 8--> %n@q where 2->1, 3->2 etc.
  # %*         gets all parms        --> %b
  # %~n        (n is 2-9) gets all parms starting with n -->%b
  #
  set pers [patternStringSearch "%" $f]
  set start 0
  set newf {}
  while {$pers != {}} {
    set pers [lassign $pers idx]
    set idx [lindex $idx 0]
    append  newf [string range $f $start $idx-1]
    set ch [string index $f $idx+1]
    frputs pers idx ch file
    switch -exact $ch {
      l -
      L -
      0 -
      1 {
	append newf $file
	incr insert
      }
      2  -
      3  -
      4  -
      5  -
      6  -
      7  -
      8  -
      9   {append newf "%[incr ch -1]@s"}
      *   {append newf "%b";frputs}
      ~   {append newf "%b"; [incr idx]}
      default {
	set pers [lassign $pers idx2]
	set idx2 [lindex $idx2 0]
	if {$idx2 == {} ||\
		! [info exists \
		       env([set ita [string toupper \
					 [string range $f $idx+1 $idx2-1]]])]} {
	  # A mal-formed env reference or something...
	  puts "Malformed string=> \"$ch\" in \"$f\". Giving up"
	  return {}
	}
	append newf [regsub "\\\\" $env($ita) {/}]
	set idx [incr idx2 -1]
      }
    }
    set start [incr idx 2]
  }
  append newf [string range $f $start end]
  if {![info exists insert]} {
    append newf " $file"
  }
  frputs f newf
  return $newf
}


# This routine fixes up a command line for MS windows.
# If possible we avoid cmd as it tends to not properly
# set pipes up and we loose the output, usually to a
# flash window.
#
# What we do is to tear the command apart a parameter at a time
# We assume that the line starts with a command so we use the
# above "windowsAutoExecOk" code to flesh it out.
# The other parameters we look at and if they look like
# filenames (testing tries to include the possibility that one
# of them may be an output file and not yet exist) we set up
# a list of such, replacing the name in the string with
# a %s (for frECF).  We then call frECF to put the whole
# command together with the names in native mode. Accepted 'args'
# are the same as frECF accepts. (In fact they are passed to
# frECF.
#
# Heck, we even do glob name expansion....
# So, parameters:
#
# cmd    is the command line. It may have %s places in it to
#        accept filename (or what ever) from:
# parms  a list of file names (or what ever) to combine into
#        the command at the %s locations (or the end). This is
#        combined with filenames generated from the command
#        tear apart in a (hopefully) intelligent manner.
# args   this is passed as 'args' to frECF (see there for
#        what they mean.
#
proc fixMSWcommand {cmd parms args} {
  set ncmd {}
  set fileList {}
  set eos [string length $cmd]
  incr eos -1
  # Tried doing this as a list but, well it was just to messy...
  # so we use a more forgiving parser...
  set nextIsCmd 1
  set last -1
  frputs cmd args
  while {
	 [set start [incr last]] < 0 ||\
	     [set last [stringGetToken $cmd start]] >= $start} {
    set part [string range $cmd $start $last]
    frputs cmd fileList part ncmd nextIsCmd
    if {[string is space [string range $cmd $last+1 end]] &&\
	    $last > $start &&\
	    [string index $part end] == "&"} {
      string replace $cmd end end " &"
      set part [string range $part 0 end-1]
      incr last -1
    }
    if {$nextIsCmd} {
      if {$ncmd == {} && $part == "exec"} {
	# command follows initial 'exec'
	set ncmd "exec"
	continue
      }
      if {[set try [windowsAutoExecOk $part]] != {}} {
	lassign $try  cm fi
	append ncmd " $cm"
	frputs cm fi
	if {$fi != {} } {
	  if {![string match *%* $try]} {
	    append ncmd " %s "
	  }
	  lappend fileList $fi
	}
	set nextIsCmd 0
	continue
      } else {
	error "Command $part not found."
      }
    }
    # Not a command. 
    set ch [string index $part 0]
    if {$ch == "|"} {
      # looks like a pipe connector
      incr nextIsCmd
      append ncmd " $ch"
      if {[string length $part] > 1} {
	string replace $cmd $start+1 end " [string range $cmd $start+1 end]"
	set last $start
      }
      continue
    }
    if {$part == "%s"} {
      set parms [lassign $parms fi]
      lappend fileList $fi
      append ncmd " %s "
      continue
    }
    # Lets assume it is an expansion pattern if it contains:
    # "?", "*", "[...]" or "{a,b,c..}"
    # The "{a,b,c..}" construct must contain the commas
    # These patterns may (usually will) reside in something that
    # otherwise looks like a file name
    #
    frputs  part ncmd
    if {[string match "*\\**" $part] ||\
	    [string match "*\\?*" $part] ||\
	    [regexp {.*\[.*\].*} $part] ||\
	    [regexp {.*\{.*,+.*\}.*} $part] } {
      frputs part ncmd
      # We found a pattern... carefully normalize and do the glob thing
      # 'file join' does not change relative paths to absolute...
      # but does flip the back slashes
      set gfile [lsort [glob -nocomplain -- [file join $part]]]
      lappend fileList {*}$gfile
      append ncmd [string repeat " %s" [llength $gfile]]
      continue
    }
    # So what does a file name look like???
    frputs "simple string or file  " part ncmd
    if {[string match -nocase "\[a-z\]:/"  $part] ||\
	    [string range $part 0 1] == "//" || \
	    [set slco [stringCount "/" $part]] > 1 ||\
	    ($ch != "/" && $slco > 0) ||\
	    [file exists $part] ||\
	    [file exists [string range [pwd] 0 1]$part]} {
      # presumed file reference...
      lappend fileList $part
      append ncmd " %s"
      frputs "looks like file  " part ncmd
    } else {
      # not a file, pass it through
      append ncmd " $part"
      frputs "looks like not a file  " part ncmd
    }
  }
  # Now put it all together with frECF
  lappend fileList {*}$parms
  return [frECF $ncmd $fileList [concat {*}$args]]
}

