# <20180510.1546.06>

# This command attempts to remove any old temp files

proc clearOldTmp {} {
  global glob env
  set mes {}
  if {[file exists $glob(conf_dir)/tmp]} {
    # still around, can we prove it is stale?
    set files [glob -directory $glob(conf_dir)/tmp/ * .*]
    set rm {}
    set bt 0
    set btl {}
    frputs files 
    if {$files != {}} {
      set rm $files
      # Lets see if we can come up with an idea of boot time
      if {! $::MSW} {
	if {[file exists /proc]} {
	  # This is close enough and quite fast if it is Linux
	  lappend btt [file mtime /proc]
	}
	if {[catch {exec uptime} utout] == 0 &&\
		      [regexp {.* up ([0-9]+) days.*} $utout duh days]} {
	  # this should work for most other *nixs (86400 seconds in a day)
	  lappend btt [expr {[clock seconds] - ($days * 86400)}]
	}
      } else {
	if {[catch {twapi::get_system_uptime} ti] == 0 && $ti > 60} {
	  lappend btl [expr {$bt - $ti}]
	} elseif {[catch {exec wmic os get lastbootuptime} btout] == 0} {
	  set btout [regsub -all {[^0-9]*|\..*$} $btout {}]
	  lappend btl [clock scan $btout -format {%Y%m%d%H%M%S}]
	  frputs btl
	} elseif {[file exists $env(SystemDrive)/TEMP]} {
	  # This seems to be useful. I think it is reset at boot time
	  # and nobody uses it (Windows/TEMP is used instead)
	  lappend btl [file mtime  $env(SystemDrive)/TEMP]
	  frputs btl
	}
      }
      foreach btt $btl {
	frputs "[clock format $btt] "
	if {$btt < $bt} {
	  set bt $btt
	}
      }
      # If we are here we have a reasonable boot time
      # Well, maybe not, but then it is 0
      set rm {}
      set mes {}
      foreach file $files {
	if {[file mtime $file] < $bt} {
	  if {[catch {file delete -force $file} ms]} {
	    lappend mes $ms
	  }
	} else {
	  lappend rm $file
	}
      }
    }
    if {$rm == {}} {
      frputs
      if {[catch {file delete -force $glob(conf_dir)/tmp} ms]} {
	lappend mes $ms
      }
    }
  }
  if {$mes != {}} {
    lappend ::mess "Error trying to remove old $glob(conf_dir)/tmp:" {*}$mes
  }
}


# The purpose of this code is to update the config file of 17.04.01.19
# to the new format. The changes are:
#
# 1.) the VFS package for adb (and thus its URL) changes from
#     'aftp' to 'adb'. This may appear in the 'Hotlist'
#
# 2.) the 'ftp' log in list (names & passwords) are moved from
#     config(ftp,login) to config(vfs,login). In addition, the
#     name of the target is now a full URL.
#
# This code moves 'ftp' logins to the new list and format insofar
# as it can. If a particular login in the 'ftp' list can not be
# figured out, it is left there with a note to the user each time
# filerunner runs to do the conversion manually.
#
proc update17.04.19 {} {
  global config
  # First fix up any "aftp" entries in the hotlist
  # We are going to cheat here and treat the hotlist as a string
  #
  set test $config(hotlist)
  set config(hotlist) [regsub -all {aftp://} $config(hotlist) {adb://}]
  set configMod [expr {$test != $config(hotlist)}]
  
  # Gosh, that was easy!

  proc getHotListEnt {hlist} {
    upvar url url
    
    foreach ent $hlist {
      lassign $ent name value
      if {$value == {}} {
	set value $name
      }
      if {[string index $name 0] != "-"} {
	if {[string match *://* $value]} {
	  if {![IsVFS $value]} {continue}
	  lappend url $VFStok
	  set url [lsort -unique $url]
	}
	continue
      }
      getHotListEnt $value
    }
  }
  set url {}
  getHotListEnt $config(hotlist)
  # Lets try to enumerate the CIFS systems.
  # This should put all we know about into the passwordLocker
  set openPassword [findUNC::enumCIFS]
  
  # now the logins
  set suggestion {}
  set result {}
  if {[info exists config(ftp,login)]} {
    foreach ent $config(ftp,login) {
      set left [lassign $ent add namePass]
      lassign $namePass name Pass
      set can ${name}@$add
      lassign [split $add .] possProt
      # set passWords(${name}@$add) $Pass
      
      # Lets see how well we do just guessing...
      # lassign [split $can @] name addr
      lassign [split $add :] addr2 port
      set found 0
      foreach ent2 $url {
	IsVFS $ent2
	lassign [split $VFSadd @] ename eadd
	if {$eadd == {}} {
	  set eadd $ename
	  set ename {}
	}
	if {$add == $eadd && ($ename == {} || $ename == $name)} {
	  if {$port != 3721 || $VFSpro == "adb" || [llength $Pass] > 4} {
	    incr found
	    lappend result  [list ${VFSpro}://$can $Pass {*}$left]
	  }
	}
      }
      # Here we dig a bit deeper...
      # We may double up here, but we will fix that with sort
      if {$possProt in {ftp sftp adb rsync}} {
	lappend result [list ${possProt}://$can $Pass {*}$left]
	incr found
      }
      set yes 0
      foreach pat {usb ser* emu* pro* con*} {
	if {[string match $pat $name]} {
	  set yes 1
	  break
	}
      }	
      if {($yes || $port == 3721) && [llength $Pass] <= 4} {
	# This looks like an adb connection...
	lappend result [list "adb://$can" $Pass {*}$left]
	incr found
      } elseif {$name == "ftp"} {
	# public ftp
	lappend result [list  "ftp://$can"  $Pass {*}$left]
	incr found
      } elseif {$found == 0} {	
	# Unknown, put it in the revised ftn,login
	lappend suggestion $ent
      }
    }
    set result [lsort -unique $result]
    set lenRemain [llength $suggestion]
    set orCount [expr {[llength $config(ftp,login)] - $lenRemain}]
    set mess0 "Filerunner has converted $orCount of your ftp login(s) to [llength $result]\
VFS login(s).
Please check the result for correctness by"
    
    set messRm "Filerunner could not convert $lenRemain ftp login(s). 
These remain in config(ftp,login). 
Please either manually convert these 
or eliminate them by"
    
    set messEd "editing the Configuration:
\"Configuration -> Edit Configuration\"
section \"VFS\",\"Vfs Login\" and \"Ftp Login\""

    set messDim "\n\nThis message is also at the beginning of the
\"Log\" which may be viewed by clicking 
\"File -> View Log...\".\n
Click this window to dismess it."

    switch [expr {[llength $result] > 0 | (($lenRemain > 0) * 2)}] {
      0 {set mes {}}
      1 {set mes "$mess0 $messEd $messDim"}
      2 {set mes "$messRm $messEd $messDim"}
      3 {set mes "$mess0 $messEd \n\n$messRm $messEd $messDim"}
    }
    set test [lnorm $config(ftp,login)]
    if {[info exist config(vfs,login)]} {
      set loginTmp $config(vfs,login)
    }
    if {[info exists loginTmp]} {
      set loginTmp [lsort -unique [concat $loginTmp $result]]
    } elseif {$result != {}} {
      set loginTmp  [lsort -unique $result]
    }
    set config(ftp,login) $suggestion
    set configMod [expr {$configMod || $test != [lnorm $config(ftp,login)]}]
    # SaveConfig no longer saves loginTmp
    # so it is really just a local tmp
  }
  # Ok, step two, move all of these to the passwordLocker.
  # Be mindfull of their might already be entries there.
  set test $config(passwordLocker)
  if {[info exists loginTmp]} {
    foreach ent $loginTmp {
      if {$ent ni $config(passwordLocker)} {
	lappend config(passwordLocker) $ent
      }
    }
  }
  # Now lets grab the passwords from the rsync table...
  set new {}
  set dPwd {}
  foreach ent $config(rsync,table) {
    lassign $ent add du duh duhh pwd
    set try {}
    if {$add == {} && $pwd == {}} {continue}
    # what sort of thing is it?
    # Use the default pwd if this entry is nil
    if {$pwd == {}} {set pwd $dPwd}
    if {[IsVFS $add]} {
      set try $add
    } elseif {[string range $add 0 1] == {//}} {
      set try "cifs:$add"
      # lappend new [list "cifs:$add" $pwd]
    } elseif {[string match *:* $add]} {
      # Lets assume this is "nfs" until we know better
      set try "nfs://$add"
    } elseif {$add == {}} {
      # this is the defaults. Save the password
      set dPwd $pwd
    }
    # Make sure we have a username in the entry
    if {[IsVFS $try]} {
      lassign  [split $VFSadd @] user ladd
      if {$ladd != {}} {
	set try $VFSpro://$::env(USER)@$ladd/$VFSpath
      }
    }
    
    if {$try != {}} {
      if {[set indx [lsearch -exact -index 0 $config(passwordLocker) $try]] == -1} {
	lappend new [list $try $pwd]
      } else {
	# make sure there was a pw
	set ext [lassign [lindex $config(passwordLocker) $indx] ad pw]
	frputs ad pw ext pwd 
	if {$pw == {} && $pwd != {}} {
	  set config(passwordLocker) [lreplace $config(passwordLocker) $indx $indx \
					  [list $ad $pwd {*}$ext]]
	}
      }
    }			      
  }

  set configMod [expr {$configMod || $config(passwordLocker) != $test}]
  
  if {[info exist openPassword] && $configMod} {
    unset openPassword
    append mes "\nFound open passwords in one or more of the files:\
             \n\"/etc/fstab\", \"/etc/mtab\" and \"/etc/auto.misc\"\
             \n\nJust saying"
  }
  if {$configMod} {
    SaveConfig
  }
  append mes [clearOldTmp]  
  # lh $result
  # lh [lsort -unique [concat [lnorm $loginTmp] $result]]
  return $mes
}
